/*	
 *	cliprend.c
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2002 Naoyuki Sawa
 *
 *	* Wed Feb 13 06:00:00 JST 2002 Naoyuki Sawa
 *	- 쐬JnB
 *	* Fri Mar  8 07:00:00 JST 2002 Naoyuki Sawa
 *	- render_line()ŏI_NbsOꂽɁA
 *	  ʒ[1sNZԂoĂ܂CB
 *	* Wed Apr 24 10:11:00 JST 2002 Naoyuki Sawa
 *	- render_polygon()ǉB
 *	* Wed Jun  5 13:51:00 JST 2002 Naoyuki Sawa
 *	- render_point()ǉB
 *	* Wed Jun 18 21:39:00 JST 2002 Naoyuki Sawa
 *	- render_fill()ǉB
 *	* Sat Jul  6 08:42:00 JST 2002 Naoyuki Sawa
 *	- surface_point()ǉB
 *	* Sat Jul  6 17:51:00 JST 2002 Naoyuki Sawa
 *	- render_fill()̊֐render_paint()ɕύXB
 *	- render_rectangle()ǉB
 *
 *	TODO:render_paint()Asurface_point()g悤ɕύXB
 */
#include "clip.h"

/*****************************************************************************
 *	
 *****************************************************************************/

TEXTURE*
texture_init(TEXTURE* texture, const void* data)
{
	TEXTUREHEADER* header;

	header = (TEXTUREHEADER*)data;

	/* wb_܂B
	 * * P/ECE̎dlɁu1bitf[^masktv(bpp=1,mask=1)݂͑܂񂪁A
	 *   {Cuł́ALȃeNX``ƂĂƂėp\Ȃ̂ŁA
	 *   G[ɂ͂܂B
	 * * eNZWƂĎw\Ȃ̂0`255ɐ܂A
	 *   eNX`̂̃TCY256hbgȏł\܂B
	 *   ]āAł̓TCY̏͌Ă܂B
	 */
	if(strncmp((char*)&header->head, "PMBP", 4) != 0) DIE();
	if(header->bpp  != 1 && header->bpp  != 2) DIE();
	if(header->mask != 0 && header->mask != 1) DIE();
	if(header->w < 8 || (header->w & 7) != 0) DIE();
	/* if(header->h < 8 || (header->h & 7) != 0) DIE(); -> CFcTCY͔CӂłB */

	/* eNX`wb_ƃ|C^i[܂B */
	memset(texture, 0, sizeof(TEXTURE));
	texture->header = *header;
	texture->buf  = (unsigned char*)(header + 1);
	if(header->mask) {
		texture->mask = (unsigned char*)(header + 1) +
				header->w * header->h / (8 / header->bpp);
	}

	return texture;
}

void
surface_clear(SURFACE* surface, int color)
{
	int n;

	n = surface->w * surface->h;
	if(surface->vbuff != NULL) memset(surface->vbuff, color, n);
	if(surface->zbuff != NULL) memset(surface->zbuff, 0xFF, n);
}

int
surface_point(SURFACE* surface, int x, int y)
{
	return 0 <= x && x <= surface->w - 1 &&
	       0 <= y && y <= surface->h - 1 ?
		surface->vbuff[surface->w * y + x] : -1;
}

/*****************************************************************************
 *	RENDER
 *****************************************************************************/

#define N_VERTICES_MAX	8	/* 1|S̍ő咸_ */

/* NbsOp_ */
typedef struct _CLIPVERTEX {
	float x, y, z;		/* + 0,12 _W       */
	unsigned short s, t;	/* +12, 4 eNX`W */
				/* =16                   */
} CLIPVERTEX;

/* `p_ */
typedef struct _RENDVERTEX {
	unsigned short x, y;	/* + 0,4 _W               */
	unsigned short z;	/* + 4,2 yl(8.8bit)           */
	unsigned short s, t;	/* + 6,4 eNX`W(8.8bit) */
				/* =10                          */
} RENDVERTEX;

/* DDA */
typedef struct _DDA {
	int n;			/* + 0,4 [̑     */
	int d;			/* + 4,4 [             */
	int u;			/* + 8,4 ̕ω     */
	int v;			/* +12,4 Jグ̕ω */
				/* =16                    */
} DDA;

static CLIPVERTEX clip_vertices[2][N_VERTICES_MAX];	/* NbsOp_obt@ */
static RENDVERTEX rend_vertices[N_VERTICES_MAX];	/* `p_obt@         */
static unsigned char fore_color;			/* OiF                     */
static unsigned char back_color;			/* wiF                     */
static void (*draw_scanline)(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);

static int  clip_z(RENDER* render, int nsrc);
static int clip_xy(RENDER* render, int nsrc);
/* yRyA */
static void select_draw_scanline_wz(RENDER* render);
static void draw_primitive_wz(RENDER* render, int n_vertices);
static void draw_scanline_wz_color(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_wz_1bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_wz_1bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_wz_2bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_wz_2bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
/* yRyAȂ */
static void select_draw_scanline_woz(RENDER* render);
static void draw_primitive_woz(RENDER* render, int n_vertices);
static void draw_scanline_woz_color(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_woz_1bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_woz_1bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_woz_2bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_woz_2bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);

void
render_init(RENDER* render, SURFACE* surface, int x, int y, int width, int height, float front, float back, float fovy)
{
	/* T[tFCXݒ肵܂B */
	render->surface = surface;

	/* r[|[g`ݒ肵܂B */
	if(width == 0) width = surface->w;
	if(height == 0) height = surface->h;
	render->left = x;
	render->top = y;
	render->right = x + width;
	render->bottom = y + height;
	if(render->left < 0 ) render->left = 0;
	if(render->top < 0 ) render->top = 0;
	if(render->right > surface->w) render->right = surface->w;
	if(render->bottom > surface->h) render->bottom = surface->h;

	/* NbsOݒ肵܂B */
	render->front = front;
	render->back = back;

	/* ReLXgX^bN܂B */
	render->context = render->context_stack;

	/* ϊs܂B */
	matrix_viewportz(&render->context->matrix, &MATRIX_I, x, y, width, height);
	matrix_perspect (&render->context->matrix, &render->context->matrix, fovy, (float)width / height, front);

	/* eNX`܂B */
	render->context->texture = NULL;
}

void
render_save(RENDER* render)
{
	/* X^bNgbv𕡐܂B */
	if(++render->context == render->context_stack + RENDER_CONTEXT_STACK_SIZE) DIE();
	memcpy(render->context, render->context - 1, sizeof(RENDERCONTEXT));
}

void
render_restore(RENDER* render)
{
	/* X^bNgbv̂Ă܂B */
	if(render->context-- == render->context_stack) DIE();
}

void
render_point(RENDER* render, int x, int y, int color)
{
	SURFACE* surface;
	unsigned char* vptr;
	int ofs;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* r[|[g`OȂA`܂B */
	if(x < render->left || render->right  <= x ||
	   y < render->top  || render->bottom <= y) return;

	/* sNZAhX߂܂B */
	ofs = surface->w * y + x;
	vptr = surface->vbuff + ofs;

	/* sNZ`܂B */
	*vptr = color;
}

/* SĂ܂BxłB
 * Q[ł́AhԂvʂ͏ȂƎv̂ŁA
 * ̂܂܂łɖ͂ȂƎv܂B
 */
void
render_paint(RENDER* render, int x0, int y0, int color)
{
/* hԂ_X^bNTCYF
 * ƂAȂ蕡GȓhԂł24xłA
 * Ô߁A]TŁA傫߂ɎĂƂɂ܂B
 * ܂傫ƁAP/ECEł̓X^bNӂ̂ŁAӂB
 */
#define RENDER_PAINT_STACK_SIZE	64

	SURFACE* surface;
	short stack[RENDER_PAINT_STACK_SIZE][2]; /* ߖ̂short */
	int i_stack, ofs, x, y, x1, x2;
	unsigned char c;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* hԂJn_r[|[gOȂA܂B */
	if(x0 < render->left || render->right  - 1 < x0) return;
	if(y0 < render->top  || render->bottom - 1 < y0) return;

	/* hԂF擾܂B */
	ofs = surface->w * y0 + x0;
	c = surface->vbuff[ofs];

	/* hԂJn_ɓhԂFȂA܂B */
	/* if(c == color) return; */
	/* ->悭lÃ`FbN͕svłB
	 *   hԂJn_ɓhԂFȂA
	 *   1. X^bNɍŏ(x0,y0)ςށB
	 *   2. while()[v̒(x0,y0)oB
	 *   3. ɓhĂ̂ŁAcontinueŔ΂B
	 *   4. X^bN͋Ȃ̂ŁAIB
	 *   ƂȂAɏI܂B
	 */

	/* hԂJn_X^bNɐς݂܂B */
	stack[0][0] = x0;
	stack[0][1] = y0;
	i_stack = 1;

	/* hԂ_ȂȂ܂... */
	while(i_stack != 0) {
		/* hԂ_o܂B */
		i_stack--;
		x0 = stack[i_stack][0];
		y0 = stack[i_stack][1];

		/* ɓhĂA΂܂B */
		ofs = surface->w * y0 + x0;
		if(surface->vbuff[ofs] == color) continue;

		/* hF̍[T܂B */
		x1 = x0;
		for(;;) {
			x1--;
			if(x1 < render->left) break; /* r[|[gO */
			ofs = surface->w * y0 + x1;
			if(surface->vbuff[ofs] != c) break; /* ʐF */
		}
		x1++; /* s߂Ă̂Ŗ߂ */

		/* hF̉E[T܂B */
		x2 = x0;
		for(;;) {
			x2++;
			if(x2 > render->right - 1) break; /* r[|[gO */
			ofs = surface->w * y0 + x2;
			if(surface->vbuff[ofs] != c) break; /* ʐF */
		}
		x2--; /* s߂Ă̂Ŗ߂ */

		/* E̊ԂhԂ܂B */
		for(x = x1; x <= x2; x++) {
			ofs = surface->w * y0 + x;
			surface->vbuff[ofs] = color;
			/* pceLCDTrans()ĂԂƁAhԂ̗lq1sNZPʂŊώ@ł܂B */
			/* pceLCDTrans() */
		}
		/* pceLCDTrans()ĂԂƁAhԂ̗lq1CPʂŊώ@ł܂B */
		/* pceLCDTrans() */

		/* ̃C̒B */
		if(y0 != render->top) {
			y = y0 - 1;
			x = x1;
			for(;;) {
				/* hF̎n܂T܂B */
				for(;;) {
					if(x > x2) break;
					ofs = surface->w * y + x;
					if(surface->vbuff[ofs] == c) break;
					x++;
				}
				if(x > x2) break; /* Ȃ */

				/* hF̏IT܂B */
				for(;;) {
					if(x > x2) break;
					ofs = surface->w * y + x;
					if(surface->vbuff[ofs] != c) break;
					x++;
				}

				/* hF̏ÏʒuX^bNɐς݂܂B */
				if(i_stack < RENDER_PAINT_STACK_SIZE) {
					stack[i_stack][0] = x - 1; /* s߂Ă̂Ŗ߂ */
					stack[i_stack][1] = y;
					i_stack++;
				}
			}
		}

		/* ̃C̒B */
		if(y0 != render->bottom - 1) {
			y = y0 + 1;
			x = x1;
			for(;;) {
				/* hF̎n܂T܂B */
				for(;;) {
					if(x > x2) break;
					ofs = surface->w * y + x;
					if(surface->vbuff[ofs] == c) break;
					x++;
				}
				if(x > x2) break; /* Ȃ */

				/* hF̏IT܂B */
				for(;;) {
					if(x > x2) break;
					ofs = surface->w * y + x;
					if(surface->vbuff[ofs] != c) break;
					x++;
				}

				/* hF̏ÏʒuX^bNɐς݂܂B */
				if(i_stack < RENDER_PAINT_STACK_SIZE) {
					stack[i_stack][0] = x - 1; /* s߂Ă̂Ŗ߂ */
					stack[i_stack][1] = y;
					i_stack++;
				}
			}
		}
	}

#undef RENDER_PAINT_STACK_SIZE
}

void
render_line(RENDER* render, int x0, int y0, int x1, int y1, int color)
{
	/* NbsOp}NB */
#define CLIP(X, Y, A, B, OP, LIM) \
	if(!((X##A) OP (LIM))) { \
		if(!((X##B) OP (LIM))) { \
			return; \
		} else { \
			(Y##A) = (Y##A) + \
			((Y##B) - (Y##A)) * ((LIM) - (X##A)) / ((X##B) - (X##A)); \
			(X##A) = (LIM); \
			clipped |= (A); /* (x1,y1)Nbvclipped=1 */ \
		} \
	}

	SURFACE* surface;
	unsigned char* vptr;
	int nx, ny;
	int ux, uy;
	int d, n;
	int tmp, ofs;
	/* I_̃NbsOtOB
	 * l0ŁA(x1,y1)Nbv1ɂȂ܂B
	 * (x0,y0)Nbvꂽǂɂ͉e܂B
	 * ͏ɁA(x0,y0)(x1,y1)Ɍĕ`܂B
	 * Az肵Ă̂ŁAʏAI_͕`܂񂪁A
	 * (x1,y1)Nbvꂽꍇ̂݁AI_`܂B
	 */
	int clipped = 0;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* e_AE㉺ŃNbv܂B */
	CLIP(x, y, 0, 1, >=, render->left);
	CLIP(x, y, 0, 1, <=, render->right - 1);
	CLIP(x, y, 1, 0, >=, render->left);
	CLIP(x, y, 1, 0, <=, render->right - 1);
	CLIP(y, x, 0, 1, >=, render->top);
	CLIP(y, x, 0, 1, <=, render->bottom - 1);
	CLIP(y, x, 1, 0, >=, render->top);
	CLIP(y, x, 1, 0, <=, render->bottom - 1);

	/* n_AhX߂܂B */
	ofs = surface->w * y0 + x0;
	vptr = surface->vbuff + ofs;

	/* XDDAp^vZ܂B */
	if(x1 >= x0) {
		nx = x1 - x0;
		ux =  1;
	} else {
		nx = x0 - x1;
		ux = -1;
	}

	/* YDDAp^vZ܂B */
	if(y1 >= y0) {
		ny = y1 - y0;
		uy =  surface->w;
	} else {
		ny = y0 - y1;
		uy = -surface->w;
	}

	/* 厲(X)A𕛎(Y)Ƃ܂B */
	if(nx < ny) {
		tmp = x0; x0 = y0; y0 = tmp;
		tmp = x1; x1 = y1; y1 = tmp;
		tmp = nx; nx = ny; ny = tmp;
		tmp = ux; ux = uy; uy = tmp;
	}
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDA[vB */
	d = -nx;
	n = nx + clipped; /* I_NbvꂽAI_`B */
	do {
		*vptr = color;
		vptr += ux;
		d += ny * 2;
		if(d >= 0) {
			vptr += uy;
			d -= nx * 2;
		}
	} while(--n); /* I_͕`Ȃ̂ŁAvfNgłB   */
		      /* I_NbvꂽAI_`܂A */
		      /* ̏ꍇ́A[vOɗ\+1Ă̂ŁA */
		      /* ɂ悱ł̓vfNgłB   */

#undef CLIP
}

void
render_rectangle(RENDER* render, int x, int y, int w, int h, int color)
{
	int x0, y0, x1, y1;

	/* TCY0x0Ȃ牽܂B */
	if(w == 0 || h == 0) return;

	/* K܂B(`Ě`) */
	if(w < 0) { x += w; w = -w; }
	if(h < 0) { y += h; h = -h; }

	/* _̏ꍇ̂݁Arender_line()ł͕`Ȃ̂ŁAʏłB */
	if(w == 1 && h == 1) {
		render_point(render, x, y, color);
		return;
	}

	/* ``܂B
	 * * ݂̎͊ȒP̂߂render_line()gĂ܂A
	 *   E`Ȃ̂ŁAƍ\łB
	 */
	x0 = x;
	y0 = y;
	x1 = x + w - 1;
	y1 = y + h - 1;
	render_line(render, x0, y0, x1, y0, color); /* `E */
	render_line(render, x1, y0, x1, y1, color); /* E`E */
	render_line(render, x1, y1, x0, y1, color); /* E` */
	render_line(render, x0, y1, x0, y0, color); /* ` */
}

/* * |S`͑SĐŏł܂A݂̎ł͂ȂĂ܂B
 *   bV`̃R[h𗬗p߁ANbsOɕgĂ܂B
 *   ̕Ǝ̐ɏ΁A啝ɍł͂łB(TODO)
 * * Ǝv܂ANbsỎǂł́A3%xȂȂ悤łB
 *   |S`ł͂yRyAsvȂ̂ŁAȗ25%xȂ܂B
 *   ̂߂ɁAyRyÃXLC`ǉ̂ǂłB
 * * ->Ă݂܂B
 */
void
render_polygon(RENDER* render, const VERTEX* vertices, int n_vertices, int color)
{
	int i;
	const VERTEX* vertex;
	CLIPVERTEX *clip_vertex;
	RENDVERTEX *rend_vertex;

	/* XLC`R[hI܂B */
	select_draw_scanline_woz(render);

	/* OiFEwiFi[܂B */
	fore_color = color >> 4 & 15;
	back_color = color      & 15;

	/* NbsOp_XggݗĂ܂B */
	if(n_vertices > N_VERTICES_MAX) n_vertices = N_VERTICES_MAX;
	for(i = 0, vertex = vertices, clip_vertex = clip_vertices[0];
	    i < n_vertices;
	    i++, vertex++, clip_vertex++) {
		clip_vertex->x = vertex->x;
		clip_vertex->y = vertex->y;
		clip_vertex->z = render->front; /* ʃNbsÔ߂̃_~[B0ȊOȂ牽łOKłB */
		clip_vertex->s = (unsigned short)(vertex->s * 0x100 + 0x80); /* +0x80ŁAeNZ̒S|Cg */
		clip_vertex->t = (unsigned short)(vertex->t * 0x100 + 0x80); /* +0x80ŁAeNZ̒S|Cg */
	}
	if(n_vertices < 3) return;

	/* ʃNbsOs܂B */
	n_vertices = clip_xy(render, n_vertices);
	if(n_vertices < 3) return;

	/* `p_XggݗĂ܂B */
	for(i = 0, clip_vertex = clip_vertices[0], rend_vertex = rend_vertices;
	    i < n_vertices;
	    i++, clip_vertex++, rend_vertex++) {
		rend_vertex->x = (unsigned short)clip_vertex->x;
		rend_vertex->y = (unsigned short)clip_vertex->y;
		/* draw_primitive_woz()rend_vertex->zgȂ̂ŁAsl̂܂܂ővłB */
		rend_vertex->s = clip_vertex->s; /* NbsOOɊ8.8bitŒ菬ɕϊς݁B */
		rend_vertex->t = clip_vertex->t; /* NbsOOɊ8.8bitŒ菬ɕϊς݁B */
	}

	/* v~eBu`܂B */
	draw_primitive_woz(render, n_vertices);
}

void
render_mesh(RENDER* render, const MESH* mesh)
{
	/* ϊςݒ_Wobt@B
	 * MESH.nv8rbgȂ̂ŁAő255_łB
	 * P/ECE̓X^bN̂ŁAstaticɔzu܂B
	 */
	static VECTOR xformed_vectors[255];

	int i, j, n_vectors, n_tcoords, n_faces, n_face_vertices, n_vertices;
	float s, xN0, yN0, x01, y01;
	VECTOR *vectors, *vector, *vector_src, *vector_dst;
	TCOORD *tcoords, *tcoord;
	FACE *face;
	FACEVERTEX *face_vertices, *face_vertex;
	CLIPVERTEX *clip_vertex;
	RENDVERTEX *rend_vertex;

	/* XLC`R[hI܂B */
	select_draw_scanline_wz(render);

	/* Xg̐擪AhXvZ܂B */
	n_vectors = mesh->nv;
	n_tcoords = mesh->nc;
	n_faces = mesh->nf;
	vectors = (VECTOR*)(mesh + 1);
	tcoords = (TCOORD*)(vectors + n_vectors);
	face = (FACE*)(tcoords + n_tcoords);

	/* S_Wϊ܂B */
	for(i = 0, vector_src = vectors, vector_dst = xformed_vectors;
	    i < n_vectors;
	    i++, vector_src++, vector_dst++) {
		vector_xform(vector_dst, &render->context->matrix, vector_src);
	}

	/* etFCX`܂B */
	for(i = 0;
	    i < n_faces;
	    i++, face = (FACE*)(face_vertices + n_face_vertices)) {
		n_face_vertices = face->nfv;
		fore_color = face->c >> 4 & 15;
		back_color = face->c      & 15;
		face_vertices = (FACEVERTEX*)(face + 1);

		/* NbsOp_XggݗĂ܂B
		 * * eNX`̏ꍇsĂ܂񂪁AɖȂ͂łB
		 *   eNX`Ȃmesh->nc=0Ȃ̂ŁAtcoords͍ŏ̃tFCXwĂ܂B
		 *   etFCX_face_vertex->ic=0Ȃ̂ŁAeNX`WƂčŏ̃tFCXm,n擾܂B
		 *   mesh->nf=0̏ꍇ̂ݍŏ̃tFCX݂܂񂪁Ȁꍇ͂̃[v̂܂B
		 *   擾eNX`W͕`掞ɖ̂ŁAeNX`Wl͊֌W܂B
		 *   L̒ʂAeNX`W擾ɕsȃ̈ɃANZXȂƂۏ؂܂B
		 *   ]āAeNX`̏ꍇsȂĂAɖ͂ȂƎv܂B
		 */
		n_vertices = 0;
		for(j = 0, face_vertex = face_vertices, clip_vertex = clip_vertices[0];
		    j < n_face_vertices;
		    j++, face_vertex++, clip_vertex++) {
			vector = &xformed_vectors[face_vertex->iv];
			tcoord = &tcoords[face_vertex->ic];
			clip_vertex->x = vector->x;
			clip_vertex->y = vector->y;
			clip_vertex->z = vector->z;
			clip_vertex->s = (unsigned short)(tcoord->s * 0x100 + 0x80); /* +0x80ŁAeNZ̒S|Cg */
			clip_vertex->t = (unsigned short)(tcoord->t * 0x100 + 0x80); /* +0x80ŁAeNZ̒S|Cg */
			if(++n_vertices == N_VERTICES_MAX) break;
		}
		if(n_vertices < 3) continue;

		/* yNbsOs܂B */
		n_vertices = clip_z(render, n_vertices);
		if(n_vertices < 3) continue;

		/* ϊ̂߂ x=x/z,y=y/z s܂B */
		for(j = 0, clip_vertex = clip_vertices[0];
		    j < n_vertices;
		    j++, clip_vertex++) {
			clip_vertex->x /= clip_vertex->z;
			clip_vertex->y /= clip_vertex->z;
		}

		/* \s܂B
		 * E(=\)LqȂ̂ŁA(=)Ȃ`܂B
		 */
		xN0 = clip_vertices[0][0].x - clip_vertices[0][n_vertices - 1].x;
		yN0 = clip_vertices[0][0].y - clip_vertices[0][n_vertices - 1].y;
		x01 = clip_vertices[0][1].x - clip_vertices[0][             0].x;
		y01 = clip_vertices[0][1].y - clip_vertices[0][             0].y;
		s = xN0 * y01 - x01 * yN0;
		if(s <= 0) continue;

		/* ʃNbsOs܂B */
		n_vertices = clip_xy(render, n_vertices);
		if(n_vertices < 3) continue;

		/* `p_XggݗĂ܂B */
		for(j = 0, clip_vertex = clip_vertices[0], rend_vertex = rend_vertices;
		    j < n_vertices;
		    j++, clip_vertex++, rend_vertex++) {
			rend_vertex->x = (unsigned short)clip_vertex->x;
			rend_vertex->y = (unsigned short)clip_vertex->y;
			rend_vertex->z = (unsigned short)((clip_vertex->z - render->front) * render->back   /
							  (render->back   - render->front) / clip_vertex->z * 0xFFFF);
			rend_vertex->s = clip_vertex->s; /* NbsOOɊ8.8bitŒ菬ɕϊς݁B */
			rend_vertex->t = clip_vertex->t; /* NbsOOɊ8.8bitŒ菬ɕϊς݁B */
		}

		/* v~eBu`܂B */
		draw_primitive_wz(render, n_vertices);
	}
}

/* yNbsOs܂B
 * [in]
 *	render		 _[ւ̃|C^B
 *	nsrc		 NbsO钸_̐B
 *	clip_vertices[0] NbsO钸_XgB(O[oϐ)
 * [out]
 *	߂l		 NbsO̒_̐B
 *	clip_vertices[0] NbsO̒_XgB(O[oϐ)
 * [note]
 *	* Ɨ̈ƂāAclip_vertices[1]g܂B
 */
static int
clip_z(RENDER* render, int nsrc)
{
#define CLIP(SRC, DST, OP, LIM) \
	src = SRC; \
	dst = DST; \
	ndst = 0; \
	lim = LIM; \
	for(i = 0; i < nsrc; i++) { \
		v0 = src +  i; \
		v1 = src + (i + 1) % nsrc; \
		if(v0->z OP lim) { \
			*dst = *v0; \
			dst++; \
			ndst++; \
			if(ndst == N_VERTICES_MAX) break; \
			if(v1->z OP lim) { \
				continue; \
			} else { \
				/* go clip */ \
			} \
		} else { \
			if(v1->z OP lim) { \
				/* go clip */ \
			} else { \
				continue; \
			} \
		} \
		/* do clip */ \
		dst->z = lim; \
		n0 = lim - v0->z; \
		n1 = v1->z - lim; \
		n01 = n0 + n1; \
		dst->x = v0->x + (v1->x - v0->x) * n0 / n01; \
		dst->y = v0->y + (v1->y - v0->y) * n0 / n01; \
		dst->s = v0->s + (v1->s - v0->s) * n0 / n01; \
		dst->t = v0->t + (v1->t - v0->t) * n0 / n01; \
		dst++; \
		ndst++; \
		if(ndst == N_VERTICES_MAX) break; \
	} \
	if(ndst < 3) return 0; \

	CLIPVERTEX *src, *dst;
	CLIPVERTEX *v0, *v1;
	float lim, n0, n1, n01;
	int i, ndst;

	CLIP(clip_vertices[0], clip_vertices[1], >=, render->front); nsrc = ndst;
	CLIP(clip_vertices[1], clip_vertices[0], <=, render->back );
	return ndst;

#undef CLIP
}

/* ʃNbsOs܂B
 * [in]
 *	render		 _[ւ̃|C^B
 *	nsrc		 NbsO钸_̐B
 *	clip_vertices[0] NbsO钸_XgB(O[oϐ)
 * [out]
 *	߂l		 NbsO̒_̐B
 *	clip_vertices[0] NbsO̒_XgB(O[oϐ)
 * [note]
 *	* Ɨ̈ƂāAclip_vertices[1]g܂B
 */
static int
clip_xy(RENDER* render, int nsrc)
{
#define CLIP(SRC, DST, X, Y, OP, LIM) \
	src = SRC; \
	dst = DST; \
	ndst = 0; \
	lim = LIM; \
	for(i = 0; i < nsrc; i++) { \
		v0 = src +  i; \
		v1 = src + (i + 1) % nsrc; \
		if(v0->X OP lim) { \
			*dst = *v0; \
			dst++; \
			ndst++; \
			if(ndst == N_VERTICES_MAX) break; \
			if(v1->X OP lim) { \
				continue; \
			} else { \
				/* go clip */ \
			} \
		} else { \
			if(v1->X OP lim) { \
				/* go clip */ \
			} else { \
				continue; \
			} \
		} \
		/* do clip */ \
		dst->X = lim; \
		n0 = lim - v0->X; \
		n1 = v1->X - lim; \
		n01 = n0 + n1; \
		dst->Y = v0->Y + (v1->Y - v0->Y) * n0 / n01; \
		n0 *= v0->z; \
		n1 *= v1->z; \
		n01 = n0 + n1; \
		dst->z = v0->z + (v1->z - v0->z) * n0 / n01; \
		dst->s = v0->s + (v1->s - v0->s) * n0 / n01; \
		dst->t = v0->t + (v1->t - v0->t) * n0 / n01; \
		dst++; \
		ndst++; \
		if(ndst == N_VERTICES_MAX) break; \
	} \
	if(ndst < 3) return 0; \

	CLIPVERTEX *src, *dst;
	CLIPVERTEX *v0, *v1;
	float lim, n0, n1, n01;
	int i, ndst;

	CLIP(clip_vertices[0], clip_vertices[1], x, y, >=, render->left  ); nsrc = ndst;
	CLIP(clip_vertices[1], clip_vertices[0], x, y, <=, render->right ); nsrc = ndst;
	CLIP(clip_vertices[0], clip_vertices[1], y, x, >=, render->top   ); nsrc = ndst;
	CLIP(clip_vertices[1], clip_vertices[0], y, x, <=, render->bottom);
	return ndst;

#undef CLIP
}

/****************************************************************************
 *	yRyA
 ****************************************************************************/

/* XLC`֐I܂B
 * [in]
 *	render		_[ւ̃|C^B
 * [out]
 *	draw_scanline	XLC`֐ւ̃|C^B(O[oϐ)
 */
static void
select_draw_scanline_wz(RENDER* render)
{
	TEXTURE* texture;

	texture = render->context->texture;
	if(texture == NULL) {
		draw_scanline = draw_scanline_wz_color;
	} else {
		if(!texture->header.mask) {
			switch(texture->header.bpp) {
			case 1:
				draw_scanline = draw_scanline_wz_1bit;
				break;
			default/*2*/:
				draw_scanline = draw_scanline_wz_2bit;
				break;
			}
		} else {
			switch(texture->header.bpp) {
			case 1:
				draw_scanline = draw_scanline_wz_1bitmask;
				break;
			default/*2*/:
				draw_scanline = draw_scanline_wz_2bitmask;
				break;
			}
		}
	}
}

/* v~eBu`܂B
 * [in]
 *	render		_[ւ̃|C^B
 *	n_vertices	_̐B
 *	rend_vertices	_XgB(O[oϐ)
 */
static void
draw_primitive_wz(RENDER* render, int n_vertices)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	side->cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		side->i.n = (v1->i - v0->i) % side->ny; \
		side->i.d = -side->ny; \
		side->i.u =  (v1->i - v0->i) / side->ny; \
		side->i.v =  1; \
	} else { \
		side->i.n = (v0->i - v1->i) % side->ny; \
		side->i.d = -side->ny; \
		side->i.u = -(v0->i - v1->i) / side->ny; \
		side->i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	side->cur.i += side->i.u; \
	side->i.d += side->i.n * 2; \
	if(side->i.d >= 0) { \
		side->cur.i += side->i.v; \
		side->i.d -= side->ny * 2; \
	}

	struct {
		int iv;		/* ݂̕ӂ̏I_CfNX            */
		int ny;		/* ݂̕ӂ̍̕ψ                */
		int n;		/* ݂̕ӂ̎c胉C              */
		RENDVERTEX cur;	/* ݒl                              */
		DDA x, z;	/* _WDDA(y͍Eʂ1Âi) */
		DDA s, t;	/* eNX`WDDA                   */
	} sides[2], *side;

	int i, itop, y, top, bottom;
	RENDVERTEX *v, *v0, *v1;

	/* ㉺[̈ʒuƁA[̒_߂܂B */
	itop = 0;
	top = bottom = rend_vertices[0].y;
	for(i = 1; i < n_vertices; i++) {
		v = &rend_vertices[i];
		if(v->y < top) {
			itop = i;
			top = v->y;
		}
		if(v->y > bottom) {
			bottom = v->y;
		}
	}
	if(top == bottom) return; /* |S͕`܂B */

	/* E̕ӂ܂B */
	for(i = 0; i < 2; i++) {
		side = &sides[i];
		side->iv = itop;
		side->n = 0; /* ɖ{̏s悤 */
	}

	/* [AeXLCɂ... */
	for(y = top; y < bottom; y++) {
		/* E̕ӂɂ... */
		for(i = 0; i < 2; i++) {
			side = &sides[i];
			/* ̕ӂɒB... */
			if(side->n == 0) {
				do {
					/* ̕ӂ̗[̒_߂܂B */
					v0 = &rend_vertices[side->iv];
					if(i == 0) {
						side->iv--;
						if(side->iv == -1) {
							side->iv = n_vertices - 1;
						}
					} else {
						side->iv++;
						if(side->iv == n_vertices) {
							side->iv = 0;
						}
					}
					v1 = &rend_vertices[side->iv];
					/* ̕ӂ̍̕ψʂ߂܂B */
					side->ny = v1->y - v0->y;
					/* V|SoAc̕`̓LZ܂B */
					if(side->ny < 0) return;
				} while(side->ny == 0); /* ȕӂ͖܂B */
				/* DDAZbgAbvB */
				SETUP(x);
				SETUP(z);
				SETUP(s);
				SETUP(t);
				/* ̕ӂ܂ł̃JE^B */
				side->n = side->ny;
			}
		}
		/* E̕ӂ̊ԂɁA`܂B */
		sides[0].cur.y = sides[1].cur.y = y;
		draw_scanline(render, &sides[0].cur, &sides[1].cur);
		/* E̕ӂɂ... */
		for(i = 0; i < 2; i++) {
			side = &sides[i];
			/* ̕ӂ܂ł̃C炵܂B */
			side->n--;
			/* DDAXebvsB */
			STEP(x)
			STEP(z)
			STEP(s)
			STEP(t)
		}
	}
	/* [̃XLC͕`܂B */

#undef SETUP
#undef STEP
}

static void
draw_scanline_wz_color(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* eNX`ȂAtFCX̔wiFŕ`܂B */
			*vptr = back_color;
			*zptr = zval;
		}
		/* obt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

static void
draw_scanline_wz_1bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = render->context->texture;

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
			 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
			 * R[hTCYƂ̌ˍŁÂőË܂B
			 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
			 */
			sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
			tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
			ofs = texture->header.w * tval + sval;
			/* 1biteNX`ȂAeNZlɂ 1:tFCX̑OiF/0:tFCX̔wiF ŕ`܂B */
			*vptr = texture->buf[ofs / 8] >> (7 - (ofs & 7)) & 1 ? fore_color : back_color;
			*zptr = zval;
		}
		/* obt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

static void
draw_scanline_wz_1bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = render->context->texture;

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
			 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
			 * R[hTCYƂ̌ˍŁÂőË܂B
			 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
			 */
			sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
			tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
			ofs = texture->header.w * tval + sval;
			if(texture->mask[ofs / 8] >> (7 - (ofs & 7)) & 1) {
				/* 1biteNX`ȂAeNZlɂ 1:tFCX̑OiF/0:tFCX̔wiF ŕ`܂B */
				*vptr = texture->buf[ofs / 8] >> (7 - (ofs & 7)) & 1 ? fore_color : back_color;
				*zptr = zval;
			}
		}
		/* obt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

static void
draw_scanline_wz_2bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = render->context->texture;

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
			 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
			 * R[hTCYƂ̌ˍŁÂőË܂B
			 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
			 */
			sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
			tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
			ofs = texture->header.w * tval + sval;
			/* 2biteNX`ȂAeNZl̂܂ܕ`܂B */
			*vptr = texture->buf[ofs / 4] >> (3 - (ofs & 3)) * 2 & 3;
			*zptr = zval;
		}
		/* obt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

static void
draw_scanline_wz_2bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = render->context->texture;

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
			 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
			 * R[hTCYƂ̌ˍŁÂőË܂B
			 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
			 */
			sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
			tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
			ofs = texture->header.w * tval + sval;
			if(texture->mask[ofs / 8] >> (7 - (ofs & 7)) & 1) {
				/* 2biteNX`ȂAeNZl̂܂ܕ`܂B */
				*vptr = texture->buf[ofs / 4] >> (3 - (ofs & 3)) * 2 & 3;
				*zptr = zval;
			}
		}
		/* obt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

/****************************************************************************
 *	yRyAȂ
 ****************************************************************************/

/* XLC`֐I܂B
 * [in]
 *	render		_[ւ̃|C^B
 * [out]
 *	draw_scanline	XLC`֐ւ̃|C^B(O[oϐ)
 */
static void
select_draw_scanline_woz(RENDER* render)
{
	TEXTURE* texture;

	texture = render->context->texture;
	if(texture == NULL) {
		draw_scanline = draw_scanline_woz_color;
	} else {
		if(!texture->header.mask) {
			switch(texture->header.bpp) {
			case 1:
				draw_scanline = draw_scanline_woz_1bit;
				break;
			default/*2*/:
				draw_scanline = draw_scanline_woz_2bit;
				break;
			}
		} else {
			switch(texture->header.bpp) {
			case 1:
				draw_scanline = draw_scanline_woz_1bitmask;
				break;
			default/*2*/:
				draw_scanline = draw_scanline_woz_2bitmask;
				break;
			}
		}
	}
}

/* v~eBu`܂B
 * [in]
 *	render		_[ւ̃|C^B
 *	n_vertices	_̐B
 *	rend_vertices	_XgB(O[oϐ)
 */
static void
draw_primitive_woz(RENDER* render, int n_vertices)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	side->cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		side->i.n = (v1->i - v0->i) % side->ny; \
		side->i.d = -side->ny; \
		side->i.u =  (v1->i - v0->i) / side->ny; \
		side->i.v =  1; \
	} else { \
		side->i.n = (v0->i - v1->i) % side->ny; \
		side->i.d = -side->ny; \
		side->i.u = -(v0->i - v1->i) / side->ny; \
		side->i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	side->cur.i += side->i.u; \
	side->i.d += side->i.n * 2; \
	if(side->i.d >= 0) { \
		side->cur.i += side->i.v; \
		side->i.d -= side->ny * 2; \
	}

	struct {
		int iv;		/* ݂̕ӂ̏I_CfNX            */
		int ny;		/* ݂̕ӂ̍̕ψ                */
		int n;		/* ݂̕ӂ̎c胉C              */
		RENDVERTEX cur;	/* ݒl                              */
		DDA x;		/* _WDDA(y͍Eʂ1Âi) */
		DDA s, t;	/* eNX`WDDA                   */
	} sides[2], *side;

	int i, itop, y, top, bottom;
	RENDVERTEX *v, *v0, *v1;

	/* ㉺[̈ʒuƁA[̒_߂܂B */
	itop = 0;
	top = bottom = rend_vertices[0].y;
	for(i = 1; i < n_vertices; i++) {
		v = &rend_vertices[i];
		if(v->y < top) {
			itop = i;
			top = v->y;
		}
		if(v->y > bottom) {
			bottom = v->y;
		}
	}
	if(top == bottom) return; /* |S͕`܂B */

	/* E̕ӂ܂B */
	for(i = 0; i < 2; i++) {
		side = &sides[i];
		side->iv = itop;
		side->n = 0; /* ɖ{̏s悤 */
	}

	/* [AeXLCɂ... */
	for(y = top; y < bottom; y++) {
		/* E̕ӂɂ... */
		for(i = 0; i < 2; i++) {
			side = &sides[i];
			/* ̕ӂɒB... */
			if(side->n == 0) {
				do {
					/* ̕ӂ̗[̒_߂܂B */
					v0 = &rend_vertices[side->iv];
					if(i == 0) {
						side->iv--;
						if(side->iv == -1) {
							side->iv = n_vertices - 1;
						}
					} else {
						side->iv++;
						if(side->iv == n_vertices) {
							side->iv = 0;
						}
					}
					v1 = &rend_vertices[side->iv];
					/* ̕ӂ̍̕ψʂ߂܂B */
					side->ny = v1->y - v0->y;
					/* V|SoAc̕`̓LZ܂B */
					if(side->ny < 0) return;
				} while(side->ny == 0); /* ȕӂ͖܂B */
				/* DDAZbgAbvB */
				SETUP(x);
				SETUP(s);
				SETUP(t);
				/* ̕ӂ܂ł̃JE^B */
				side->n = side->ny;
			}
		}
		/* E̕ӂ̊ԂɁA`܂B */
		sides[0].cur.y = sides[1].cur.y = y;
		draw_scanline(render, &sides[0].cur, &sides[1].cur);
		/* E̕ӂɂ... */
		for(i = 0; i < 2; i++) {
			side = &sides[i];
			/* ̕ӂ܂ł̃C炵܂B */
			side->n--;
			/* DDAXebvsB */
			STEP(x)
			STEP(s)
			STEP(t)
		}
	}
	/* [̃XLC͕`܂B */

#undef SETUP
#undef STEP
}

static void
draw_scanline_woz_color(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	unsigned char* vptr;	/* J[obt@|C^     */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	const RENDVERTEX* tmp;
	int ofs;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDA[vB */
	n = nx;
	do {
		/* eNX`ȂAtFCX̔wiFŕ`܂B */
		*vptr = back_color;
		/* obt@|C^i߂܂B */
		vptr++;
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

static void
draw_scanline_woz_1bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = render->context->texture;

	/* DDA[vB */
	n = nx;
	do {
		/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
		 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
		 * R[hTCYƂ̌ˍŁÂőË܂B
		 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
		 */
		sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
		tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
		ofs = texture->header.w * tval + sval;
		/* 1biteNX`ȂAeNZlɂ 1:tFCX̑OiF/0:tFCX̔wiF ŕ`܂B */
		*vptr = texture->buf[ofs / 8] >> (7 - (ofs & 7)) & 1 ? fore_color : back_color;
		/* obt@|C^i߂܂B */
		vptr++;
		/* DDAXebvsB */
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

static void
draw_scanline_woz_1bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = render->context->texture;

	/* DDA[vB */
	n = nx;
	do {
		/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
		 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
		 * R[hTCYƂ̌ˍŁÂőË܂B
		 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
		 */
		sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
		tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
		ofs = texture->header.w * tval + sval;
		if(texture->mask[ofs / 8] >> (7 - (ofs & 7)) & 1) {
			/* 1biteNX`ȂAeNZlɂ 1:tFCX̑OiF/0:tFCX̔wiF ŕ`܂B */
			*vptr = texture->buf[ofs / 8] >> (7 - (ofs & 7)) & 1 ? fore_color : back_color;
		}
		/* obt@|C^i߂܂B */
		vptr++;
		/* DDAXebvsB */
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

static void
draw_scanline_woz_2bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = render->context->texture;

	/* DDA[vB */
	n = nx;
	do {
		/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
		 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
		 * R[hTCYƂ̌ˍŁÂőË܂B
		 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
		 */
		sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
		tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
		ofs = texture->header.w * tval + sval;
		/* 2biteNX`ȂAeNZl̂܂ܕ`܂B */
		*vptr = texture->buf[ofs / 4] >> (3 - (ofs & 3)) * 2 & 3;
		/* obt@|C^i߂܂B */
		vptr++;
		/* DDAXebvsB */
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

static void
draw_scanline_woz_2bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = render->context->texture;

	/* DDA[vB */
	n = nx;
	do {
		/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
		 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
		 * R[hTCYƂ̌ˍŁÂőË܂B
		 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
		 */
		sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
		tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
		ofs = texture->header.w * tval + sval;
		if(texture->mask[ofs / 8] >> (7 - (ofs & 7)) & 1) {
			/* 2biteNX`ȂAeNZl̂܂ܕ`܂B */
			*vptr = texture->buf[ofs / 4] >> (3 - (ofs & 3)) * 2 & 3;
		}
		/* obt@|C^i߂܂B */
		vptr++;
		/* DDAXebvsB */
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}

